﻿'use strict';

// view model
function ChatSessionViewModel(userProviderKey, sessionId, mediaParams, guestParams, chatProfile) {
    var REPEAT_TIMEOUT = 30000, REPEAT_SESSION_CHECK = 300000;
    var DELAYED_MESSAGE_INTERVAL = 1000, MAX_EXPECTED_DELAYED_MESSAGE_INTERVAL = 5500;
    var chatHub;
    var pc = null;
    var self = this;
    var incompleteMessageLength = 0, sendingIncompleteMessage = false;
    var sessionStorage;
    
    self.sessionId = sessionId;
    self.isOnServer = ko.observable(false);
    self.userProviderKey = userProviderKey;
    self.isGuest = guestParams !== null && guestParams !== undefined;
    self.isConnected = ko.observable(false);
    self.isLoading = ko.observable(false);
    self.chatProfileOperatorsAvailability = {
        canChatWithOperator: ko.observable(null),
        canReceiveNewGuests: ko.pureComputed(function () {
            return isInGuestChatBotMode() ? true : self.chatProfileOperatorsAvailability.canChatWithOperator();
        }, this),
        queuedSessionsCount: ko.observable(null)
    };

    self.userSessions = ko.observableArray([]);
    self.myUserSession = ko.observable();
    self.guestUserSession = ko.observable();
    self.session = ko.observable();
    self.clientMedia = ko.observable();
    self.userAgent = new UAParser().getResult();
    self.securedConnection = window.location.protocol === "https:";

    self.attachmentsUpload = ko.observable(null);
    self.attachmentsUploadError = ko.observable(null);
    self.messageToSend = ko.observable("");
    self.delayedMessageToSend = ko.pureComputed(self.messageToSend).extend({ rateLimit: DELAYED_MESSAGE_INTERVAL });
    self.delayedMessageToSend.subscribe(function (messageText) {
        if ((messageText.length > 0 || incompleteMessageLength !== null) && !sendingIncompleteMessage && chatHub && self.isOnServer()) {
            sendingIncompleteMessage = true;
            chatHub.server.addMessage(guestParams ? messageText : null, sessionId, messageText.length).done(function () {
                sendingIncompleteMessage = false;
            }).fail(function (error) {
                sendingIncompleteMessage = false;
                console.warn('sendingIncompleteMessage: ' + error);
            });

            incompleteMessageLength = messageText.length;
        }
    });

    self.sendingMessage = ko.observable(false);
    self.messages = ko.observableArray();
    self.localCameraStream = ko.observable(null);
    self.remoteCameraStream = ko.observable(null);
    self.remoteScreenStream = ko.observable(null);

    self.isLocalVideo = ko.pureComputed(function () {
        return isVideoStream(self.localCameraStream());
    }, this);

    self.isRemoteVideo = ko.pureComputed(function () {
        return isVideoStream(self.remoteCameraStream());
    }, this);

    self.isRemoteScreen = ko.pureComputed(function () {
        return self.remoteScreenStream() !== null;
    }, this);

    function isVideoStream(stream) {
        return stream instanceof MediaStream ? stream.getVideoTracks().length > 0 : false;
    }

    self.lastMessageSentDate = ko.observable();
    self.lastMessageDate = ko.observable();
    self.lastUnspottedMessage = ko.observable();
    self.unspottedEventsCounter = ko.computed(function () {
        var res = 0;
        var lastUnspottedSet = false;
        if (self.messages().length > 0) {
            for (var i = self.messages().length - 1; i >= 0; --i) {
                let message = self.messages()[i];
                if (i === self.messages().length - 1)
                    self.lastMessageDate(message.Date);

                if (message.UserProviderKey === userProviderKey) {
                    self.lastMessageSentDate(message.Date);
                    break;
                }

                if (message.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE)
                    continue;

                if (self.myUserSession()) {
                    if ((message.Type === chatUtils.ChatMessageTypeEnum.COMMON || message.Type === chatUtils.ChatMessageTypeEnum.ATTACHMENTS) &&
                        new Date(message.Date) >= new Date(self.myUserSession().SpotTimestamp)) {

                        if (!lastUnspottedSet) {
                            lastUnspottedSet = true;
                            self.lastUnspottedMessage(message);
                        }

                        ++res;
                    }
                    else break;
                }
            }
        }
        else {
            self.lastMessageDate(null);
            self.lastMessageSentDate(null);
        }

        if (res === 0)
            self.lastUnspottedMessage(null);

        return res;
    }, this);

    var myLatestSpotMessageId = 0;
    self.spot = function () {
        if (!chatHub)
            return;

        for (var i = self.messages().length - 1; i >= 0; i--) {
            var message = self.messages()[i];
            if (message.UserProviderKey !== userProviderKey &&
                (message.Type === chatUtils.ChatMessageTypeEnum.COMMON || message.Type === chatUtils.ChatMessageTypeEnum.ATTACHMENTS)) {
                if (message.Id > myLatestSpotMessageId) {
                    myLatestSpotMessageId = message.Id;
                    chatHub.server.spotUserSession(sessionId, message.Id).fail(function (error) {
                        console.warn('spotUserSessions: ' + error);
                    });
                }
                
                break;
            }
        }
    };

    self.structurizedSpots = {
        minSpottedMessageId: 0,
        spots: {}
    };

    function setSpotOfMessage(message) {
        message.spot = ko.pureComputed(function () {
            if (this.Id < self.structurizedSpots.minSpottedMessageId)
                return null;

            var res = null;
            for (var prop in self.structurizedSpots.spots) {
                var spot = self.structurizedSpots.spots[prop]();
                if (spot.messageId === this.Id) {
                    if (!res)
                        res = [];

                    res.push(spot);
                }
            }

            return res;
        }, message);
    }

    function setDeliveryOfMessage(message) {
        if (message.UserProviderKey === userProviderKey &&
            (message.Type === chatUtils.ChatMessageTypeEnum.COMMON || message.Type === chatUtils.ChatMessageTypeEnum.ATTACHMENTS)) {
            message.delivery = ko.pureComputed(function () {
                return this.Id === myLatestMessageId() && !this.spot();
            }, message);

            message.deliverySub = message.delivery.subscribe(function (newDelivery) {
                if (!newDelivery) {
                    this.deliverySub.dispose();
                    delete this.deliverySub;
                    this.delivery.dispose();
                }
            }, message);

            if (message.Id > myLatestMessageId()) {
                myLatestMessageId(message.Id);
            }
        }
    }

    var MAX_SPOTIFIED_MESSAGES = 50;
    var myLatestMessageId = ko.observable(0);
    function spotifyMessages() {
        var messagesLength = self.messages().length;
        if (messagesLength === 0)
            return;

        var i = messagesLength - 1;
        var deliverySet = false;
        do {
            var message = self.messages()[i--];
            if (message.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE)
                continue;

            if (!('spot' in message)) {
                setSpotOfMessage(message);

                if (!deliverySet) {
                    deliverySet = true;
                    setDeliveryOfMessage(message);
                }
            }
            else
                break;
        } while (i >= 0 && i >= messagesLength - MAX_SPOTIFIED_MESSAGES &&
            (message.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE || message.Id > self.structurizedSpots.minSpottedMessageId));
    }

    function structurizeSpot(userSession) {
        if (userSession.User.UserProviderKey === userProviderKey)
            throw "Spot must be set for other users only.";

        var spot = {
            userProviderKey: userSession.User.UserProviderKey,
            fullName: userSession.User.FullName,
            timestamp: new Date(userSession.SpotTimestamp),
            messageId: userSession.SpotMessageId
        };

        if (!self.structurizedSpots.spots[userSession.User.UserProviderKey])
            self.structurizedSpots.spots[userSession.User.UserProviderKey] = ko.observable(spot);
        else
            self.structurizedSpots.spots[userSession.User.UserProviderKey](spot);

        var minSpottedMessageId = Number.MAX_VALUE;
        for (var prop in self.structurizedSpots.spots) {
            var currentMessageId = self.structurizedSpots.spots[prop]().messageId;
            if (currentMessageId < minSpottedMessageId)
                minSpottedMessageId = currentMessageId;
        }

        self.structurizedSpots.minSpottedMessageId = minSpottedMessageId;
    }

    self.chatProfileKey = guestParams && guestParams.chatProfileKey || chatProfile && chatProfile.Key;
    self.chatProfile = chatProfile;
    self.chatProfileCallOptions = ko.observable();
    self.form = ko.observable();
    self.appearance = ko.observable({
        OperatorMessageBackgroundColor: "#ECECEC",
        OperatorMessageTextColor: "#000000",
        GuestMessageBackgroundColor: "#FBFBFB",
        GuestMessageTextColor: "#000000",
        BorderRadius: 0,
        ChatBot: null
    });
    self.offlineFormEnabled = null;
    self.answers = ko.observable();

    var chatProfileDataProcessed = false;
    function loadChatProfileData() {
        return new Promise(function (resolve, reject) {
            if (chatProfile) {
                processChatProfileData(chatProfile);
                resolve(chatProfile);
            }
            else if (self.chatProfileKey) {
                var chatProfileUrl = window.location.protocol + '//' + window.location.host + RQ_URL + "/Chat/GetChatProfile/" + self.chatProfileKey;
                $.getJSON(chatProfileUrl).done(function (res) {
                    if (!res) {
                        reject();
                        return;
                    }

                    self.chatProfile = res;
                    processChatProfileData(res);
                    resolve(res);
                }).fail(function (error) {
                    reject(error);
                });
            }
            else
                resolve(null);
        });
    }

    function processChatProfileData(chatProfile) {
        self.appearance({
            OperatorMessageBackgroundColor: chatProfile.OperatorMessageBackgroundColor,
            OperatorMessageTextColor: chatProfile.OperatorMessageTextColor,
            GuestMessageBackgroundColor: chatProfile.GuestMessageBackgroundColor,
            GuestMessageTextColor: chatProfile.GuestMessageTextColor,
            BorderRadius: chatProfile.BorderRadius,
            ChatBot: chatProfile.ChatBot
        });

        if (chatProfile.Form) {
            self.form({
                customFormFields: ko.observableArray(chatProfile.Form.CustomFormFields.map(function (f) {
                    return new CustomFormField(f);
                })),
                nameFormField: new SystemFormField(chatProfile.Form.NameFormField),
                emailFormField: new SystemFormField(chatProfile.Form.EmailFormField),
                telephoneFormField: new SystemFormField(chatProfile.Form.TelephoneFormField),
                formFieldType2TemplateName: function (formField) {
                    switch (formField.type) {
                        case chatUtils.CustomFormFieldTypeEnum.STRING: return "stringFormFieldTemplate";
                        case chatUtils.CustomFormFieldTypeEnum.BOOLEAN: return "booleanFormFieldTemplate";
                        default: throw { message: "Unknown form field type." };
                    }
                }
            });

            self.form().emailFormField.value.extend({ email: true });
        }

        if (chatProfile.Answers)
            self.answers(chatProfile.Answers);

        self.chatProfileCallOptions(chatProfile.CallOption);
        self.offlineFormEnabled = chatProfile.OfflineFormEnabled;
        chatProfileDataProcessed = true;
    }

    self.offlineForm = {
        name: ko.observable().extend({ required: true }),
        email: ko.observable().extend({
            required: true,
            email: true
        }),
        text: ko.observable().extend({ required: true })
    };
    self.offlineFormSent = ko.observable(false);
    self.offlineFormSendingError = ko.observable(null);

    self.formEditMode = ko.observable(false);
    self.formValuesCopy = null;
    self.toggleFormEditMode = function () {
        let form = self.form();
        if (self.formValuesCopy === null)
        {
            self.formValuesCopy = {
                name: form.nameFormField.value(),
                email: form.emailFormField.value(),
                tel: form.telephoneFormField.value(),
                custom: $.map(form.customFormFields(), function (customFormField) {
                    return customFormField.value();
                })
            };

            self.formEditMode(true);
        }
        else
        {
            form.nameFormField.value(self.formValuesCopy.name);
            form.emailFormField.value(self.formValuesCopy.email);
            form.telephoneFormField.value(self.formValuesCopy.tel);

            for (let i = 0; i < form.customFormFields().length; i++)
                form.customFormFields()[i].value(self.formValuesCopy.custom[i]);

            self.formValuesCopy = null;
            self.formEditMode(false);
        }
    };

    //functions
    function CustomFormField(f) {
        var self = this;
        self.id = f.Id;
        self.title = f.Title;
        self.required = f.Required;
        self.type = f.Type;
        self.value = ko.observable(null).extend({ required: f.Required /*&& f.Type !== chatUtils.CustomFormFieldTypeEnum.BOOLEAN*/ });
    }

    function SystemFormField(systemFormField) {
        var sff = this;
        var isUnknownGuest = userProviderKey.startsWith("guest") || !guestParams;
        sff.displayed = isUnknownGuest && systemFormField.Displayed;
        sff.required = isUnknownGuest && systemFormField.Required;
        sff.value = ko.observable(null).extend({ required: isUnknownGuest && systemFormField.Required });
    }

    self.processOfflineForm = function () {
        if (!guestParams)
            throw { error: "Form values are intended to be used with guest only." };

        if (!self.offlineFormEnabled)
            return;

        var result = ko.validation.group(self.offlineForm, { deep: false });
        if (result().length > 0) {
            result.showAllMessages(true);
        }
        else {
            if (geoStr === undefined) {
                chatUtils.requestAndSaveGeo().then(function (geo) {
                    geoStr = JSON.stringify(geo);
                    sendOfflineForm();
                }).catch(function () {
                    sendOfflineForm();
                });
            }
            else
                sendOfflineForm();
        }
    };

    function sendOfflineForm() {
        self.offlineFormSendingError(null);
        self.offlineFormSent(null);
        $.post(window.location.protocol + '//' + window.location.host + RQ_URL + "/Chat/SendOfflineForm/" + self.chatProfileKey, {
            sessionId: self.isOnServer() ? sessionId : null,
            name: self.offlineForm.name(),
            fromAddress: self.offlineForm.email(),
            text: self.offlineForm.text(),
            geo: geoStr ? JSON.parse(geoStr) : null,
            hostUrl: guestParams && guestParams.hostUrl,
            hostDocumentTitle: guestParams && guestParams.hostDocumentTitle
        }).done(function (result) {
            self.offlineFormSendingError(result === "OK" ? null : result);
            self.offlineFormSent(true);
        }).fail(function () {
            self.offlineFormSendingError("Sending offline form error.");
            self.offlineFormSent(true);
        });
    }

    self.offlineFormSent.subscribe(function (sent) {
        if (sent && self.offlineFormSendingError() === null) {
            self.offlineForm.text("");
        }
    });

    function startSessionIfNeeded() {
        return new Promise(function (resolve, reject) {
            if (!chatHub) {
                initConnectionHub();
                $.connection.hub.start().done(function () {
                    connectionDone().then(function (e) {
                        self.isOnServer(true);
                        resolve();
                    });
                });
            } else if (!self.isOnServer()) {
                chatHub.server.createGuestUserSessionIfNeeded(
                    sessionId,
                    guestParams ? guestParams.chatProfileKey : undefined,
                    clientMessagesScaffolds ? JSON.stringify(clientMessagesScaffolds) : null,
                    self.messageToSend()).done(function () {
                        self.isOnServer(true);
                        self.messageToSend(""),
                        resolve();
                    }).fail(function (error) {
                        console.warn('createGuestUserSessionIfNeeded: ' + error);
                        reject();
                    });
            }
            else
                resolve();
        });
    }

    function saveFormByHub(formVals) {
        chatHub.server.setFormValues(formVals, sessionId).done(function () {
            self.formValuesCopy = null;
            self.formEditMode(false);
        }).fail(function (error) {
            console.warn('saveFormValues: ' + error);
        });
    }

    self.validateAndSaveForm = function () {
        var result = ko.validation.group(self.form, { deep: true });
        if (result().length > 0) {
            result.showAllMessages(true);
        }
        else {
            let formVals = {
                chatProfileKey: self.chatProfileKey,
                name: self.form().nameFormField.value(),
                email: self.form().emailFormField.value(),
                telephone: self.form().telephoneFormField.value(),
                customFields: self.form().customFormFields().map(function (customFormField) {
                    return {
                        id: customFormField.id,
                        type: customFormField.type,
                        val: customFormField.value()
                    };
                })
            };

            if (formVals.name && formVals.name.length > 0)
                self.offlineForm.name(formVals.name);
            if (formVals.email && formVals.email.length > 0)
                self.offlineForm.email(formVals.email);

            startSessionIfNeeded().then(function () { saveFormByHub(formVals); });
        }
    };

    self.addWelcomeMessageIfNeeded = function () {
        if (texts && self.messages().length === 0 && (!self.form() || self.session() && self.session().Form)) {
            let useBotWelcome = isInGuestChatBotMode() && self.chatProfile.ChatBot.WelcomeText;
            let nameFormFieldValue = self.form() && self.form().nameFormField.value();
            let welcomeText = !useBotWelcome ? texts.ChatFirstMessage.replace("{visitorName}", nameFormFieldValue && nameFormFieldValue.length > 0 ? nameFormFieldValue : "") : self.chatProfile.ChatBot.WelcomeText;
            let welcomeDate = new Date().getTime();
            let clientMessageScaffold = {
                type: useBotWelcome ? chatUtils.ChatMessageTypeEnum.BOT_COMMON : chatUtils.ChatMessageTypeEnum.AUTOMATION,
                date: welcomeDate,
                text: welcomeText,
                fullName: useBotWelcome ? self.chatProfile.ChatBot.Name : ""
            };
            clientMessagesScaffolds.push(clientMessageScaffold);

            if (self.isOnServer()) {
                chatHub.server.triggerClientMessage(sessionId, clientMessageScaffold).fail(function (error) {
                    console.warn('triggerClientMessage: ' + error);
                });
            }
            else {
                placeMessageAtPosition({
                    Text: welcomeText,
                    Date: welcomeDate,
                    UserProviderKey: "",
                    FullName: useBotWelcome ? self.chatProfile.ChatBot.Name : "",
                    FromOperator: true,
                    Type: useBotWelcome ? chatUtils.ChatMessageTypeEnum.BOT_COMMON : chatUtils.ChatMessageTypeEnum.AUTOMATION,
                    ChatBot: !!useBotWelcome
                });
            }
        }
    };

    function fillFormValues(formVals) {
        let form = self.form();
        if (form && formVals) {
            if (formVals.Name)
                form.nameFormField.value(formVals.Name);

            if (formVals.Email)
                form.emailFormField.value(formVals.Email);

            if (formVals.Telephone)
                form.telephoneFormField.value(formVals.Telephone);

            if (formVals.CustomFields) {
                $(formVals.CustomFields).each(function (valIndex, fieldValue) {
                    $(form.customFormFields()).each(function (fieldIndex, formField) {
                        if (fieldValue.Id === formField.id) {
                            formField.value(fieldValue.Val);
                            return false;
                        }
                    });
                });
            }
        }
    }

    function determineClientsWebRTCCompatibility(myUserAgent, oppositeUserAgent) {
        var res = {
            isAudioCompatible: false,
            isVideoCompatible: false,
        };

        if (getAudioVideoOS_Availability(myUserAgent.os.name) && getAudioVideoOS_Availability(oppositeUserAgent.OS.name)) {
            let myBrowserAudioVideoAvailability = getBrowserAudioVideoAvailability(myUserAgent.browser);
            let oppositeBrowserAudioVideoAvailability = getBrowserAudioVideoAvailability(oppositeUserAgent.Browser);

            res.isAudioCompatible = myBrowserAudioVideoAvailability && myBrowserAudioVideoAvailability === oppositeBrowserAudioVideoAvailability;
            res.isVideoCompatible = myBrowserAudioVideoAvailability && myBrowserAudioVideoAvailability === oppositeBrowserAudioVideoAvailability;
        }

        return res;
    }

    function getAudioVideoOS_Availability(os) {
        return os.startsWith("Windows") || os === "Android" || os === "Mac OS" || os === "Firefox OS" || os === "Chromium OS";
    }

    function getScreenOS_Availability(os) {
        return os.startsWith("Windows") || os === "Mac OS" || os === "Chromium OS";
    }

    function getBrowserAudioVideoAvailability(browser) {
        var majorVer = 0;
        if (browser.version) {
            var hlpVer = parseInt(browser.version);
            majorVer = isNaN(hlpVer) ? 0 : hlpVer;
        }

        switch (browser.name) {
            case "Firefox":
                return majorVer >= 59 ? 1 : 0;
            case "Chrome":
            case "Chromium":
                return majorVer >= 64 ? 1 : 0;
            case "Opera":
                return majorVer >= 51 ? 1 : 0;
            case "Edge":
                return majorVer >= 18 ? 1 : 0;
            case "Safari":
                return majorVer >= 11 ? 1 : 0;
            default:
                return 0;
        }
    }

    function getBrowserScreenAvailability(browser) {
        var majorVer = 0;
        if (browser.version) {
            var hlpVer = parseInt(browser.version);
            majorVer = isNaN(hlpVer) ? 0 : hlpVer;
        }

        switch (browser.name) {
            case "Firefox":
                return majorVer >= 66 ? 1 : 0;
            case "Chrome":
            case "Chromium":
                return majorVer >= 72 ? 1 : 0;
            case "Opera":
                return majorVer >= 60 ? 1 : 0;
            case "Safari":
                return majorVer >= 13 ? 1 : 0;
            default:
                return 0;
        }
    }

    function getGuestFirstMessageDate() {
        if (self.messages() && self.guestUserSession())
        {
            for (let i = 0; i < self.messages().length; ++i) {
                let message = self.messages()[i];
                if (message.UserProviderKey === self.guestUserSession().User.UserProviderKey)
                    return new Date(message.Date);
            }
        }
        
        return null;
    }

    function checkAutomation(automations) {
        function evaluateNumericOperands(operand1, operand2, operator) {
            if (isNaN(operand1) || isNaN(operand2))
                return undefined;

            switch (operator) {
                case chatUtils.ChatAutomationEventConditionOperatorEnum.IS: return operand1 === operand2;
                case chatUtils.ChatAutomationEventConditionOperatorEnum.IS_NOT: return operand1 !== operand2;
                case chatUtils.ChatAutomationEventConditionOperatorEnum.IS_GREATER_THAN: return operand1 > operand2;
                case chatUtils.ChatAutomationEventConditionOperatorEnum.IS_LESS_THAN: return operand1 < operand2;
            }
        }

        function evaluateCondition(condition, type) {
            switch (condition.ConditionField) {
                case chatUtils.ChatAutomationEventConditionFieldEnum.VISITOR_IS_ON_WEB:
                    return type === chatUtils.ChatProfileAutomationEventEventTypeEnum.LoadChatWidget ?
                        evaluateNumericOperands(initIntervalSec, condition.Values[0].ValueInt, condition.ConditionOperator) :
                        firstMessageIntervalSec ? evaluateNumericOperands(firstMessageIntervalSec, condition.Values[0].ValueInt, condition.ConditionOperator) : false;
                case chatUtils.ChatAutomationEventConditionFieldEnum.VISITOR_IS_CHATTING_WITH_OPERATOR:
                    switch (condition.ConditionOperator) {
                        case chatUtils.ChatAutomationEventConditionOperatorEnum.YES: return isAcceptedByAnybody;
                        case chatUtils.ChatAutomationEventConditionOperatorEnum.NO: return !isAcceptedByAnybody;
                    }
                    break;
                case chatUtils.ChatAutomationEventConditionFieldEnum.VISITOR_SENT_MESSAGE:
                    switch (condition.ConditionOperator) {
                        case chatUtils.ChatAutomationEventConditionOperatorEnum.YES: return guestFirstMessageDate !== null;
                        case chatUtils.ChatAutomationEventConditionOperatorEnum.NO: return guestFirstMessageDate === null;
                    }
                    break;
                case chatUtils.ChatAutomationEventConditionFieldEnum.NUMBER_OF_VISITS:
                    return evaluateNumericOperands(visitsCount, condition.Values[0].ValueInt, condition.ConditionOperator);
            }
        }

        function evaluateRelatedConditions(conditions, type) {
            if (conditions.length > 0) {
                let orResult = false, andResult = true;
                for (let i = 0; i < conditions.length; ++i) {
                    let condition = conditions[i];
                    if (!condition.IsAnd && i > 0) {
                        orResult = orResult || andResult;
                        andResult = true;
                    }

                    andResult = andResult && evaluateCondition(condition, type);
                }

                return orResult || andResult;
            }
            else {
                return type === chatUtils.ChatProfileAutomationEventEventTypeEnum.SendMessageToChat ? firstMessageIntervalSec !== null : true;
            }
        }

        var isAcceptedByAnybody = null;
        if (self.chatProfileOperatorsAvailability.canReceiveNewGuests() || (isAcceptedByAnybody = self.isAcceptedByAnybody())) {
            var nowDate = new Date();
            var initIntervalSec = Math.trunc((nowDate - automationInitDate) / 1000);
            var guestFirstMessageDate = getGuestFirstMessageDate();
            var firstMessageIntervalSec = guestFirstMessageDate ? Math.trunc((nowDate - guestFirstMessageDate) / 1000) : null;
            if (isAcceptedByAnybody === null)
                isAcceptedByAnybody = self.isAcceptedByAnybody();

            $(automations).each(function (index, automation) {
                if (triggeredAutomations[automation.Id.toString()])
                    return true;

                let conditionsResult = evaluateRelatedConditions(automation.EventConditions, automation.EventType);
                if (conditionsResult && Array.isArray(automation.Actions) && automation.Actions.length > 0) {
                    let automationTime = new Date().getTime();
                    let clientMessageScaffold = {
                        type: chatUtils.ChatMessageTypeEnum.AUTOMATION,
                        date: automationTime,
                        text: automation.Actions[0].Text,
                        fullName: ""
                    };
                    clientMessagesScaffolds.push(clientMessageScaffold);
                    triggeredAutomations[automation.Id.toString()] = automationTime;
                    chatUtils.sessionStorageSetTriggeredAutomation(self.chatProfileKey, automation.Id, automationTime);

                    if (self.isOnServer()) {
                        chatHub.server.triggerClientMessage(sessionId, clientMessageScaffold).fail(function (error) {
                            console.warn('triggerAutomation: ' + error);
                        });
                    }
                    else {
                        placeMessageAtPosition({
                            Text: automation.Actions[0].Text,
                            Date: automationTime,
                            UserProviderKey: "",
                            FullName: "",
                            FromOperator: true,
                            Type: chatUtils.ChatMessageTypeEnum.AUTOMATION
                        });
                        spotifyMessages();
                        postData({ displayChat: true, sessionId: sessionId });
                    }
                }
            });
        }
    }

    var automationInitDate, triggeredAutomations, clientMessagesScaffolds = [], visitsCount;
    function startAutomationTimer(automations) {
        triggeredAutomations = sessionStorage.triggeredAutomations;
        visitsCount = sessionStorage.visitsCount;
        automationInitDate = new Date();

        setInterval(function () {
            checkAutomation(automations);
        }, 5000);
    }

    function getAndSendGeoIfNeeded() {
        if (geoStr === undefined) {
            geoStr = null;
            chatUtils.requestAndSaveGeo().then(function (geo) {
                geoStr = JSON.stringify(geo);
                chatHub.server.setGeo(geo).fail(function (error) {
                    geoStr = undefined;
                    console.warn('setGeo: ' + error);
                });
            }).catch(function () {
                geoStr = undefined;
                console.warn('requestAndSaveGeo error occured');
            });   
        }
    }

    self.setUserSessions = function (userSessions) {
        self.isOnServer(true);
        $(userSessions).each(function (index, userSession) {
            self.setUserSession(userSession);
        });

        spotifyMessages();
        getAndSendGeoIfNeeded();
    };

    self.setUserSession = function (userSession) {
        self.session(userSession.Session);
        if (userSession.Session.GuestProviderKey === userSession.User.UserProviderKey) {
            self.guestUserSession(userSession);
            fillFormValues(userSession.Session.Form);
        }

        if (userSession.User.UserProviderKey === userProviderKey) {
            self.myUserSession(userSession);
            if (userSession.Messages)
            {
                self.messages.removeAll();
                ko.utils.arrayPushAll(self.messages, userSession.Messages);
            }

            setChatBotIncompleteMessageIfNeeded(true);

            if (guestParams && userSession.Session.IsTerminated) {
                chatUtils.resetSession(self.chatProfileKey);
                postData({ terminate: true, chatProfileKey: self.chatProfileKey });
            }
        }
        else {
            let replaced = false;
            for (var i = 0; i < self.userSessions().length; i++) {
                if (self.userSessions()[i].us.User.UserProviderKey === userSession.User.UserProviderKey) {
                    self.userSessions.replace(self.userSessions()[i], GetUserSessionWithCompatibility(userSession));
                    replaced = true;
                    break;
                }
            }

            if (!replaced) {
                self.userSessions.push(GetUserSessionWithCompatibility(userSession));
            }

            structurizeSpot(userSession);
        }
    };

    self.getUserSession = function (userProviderKey) {
        for (var i = 0; i < self.userSessions().length; i++) {
            if (self.userSessions()[i].us.User.UserProviderKey === userProviderKey) {
                return self.userSessions()[i].us;
            }
        }

        return null;
    };

    self.oppositeCallingUserSession = function () { return self._oppositeCallingUserSessionGetter.apply(self); };
    self.chatMediaSessionParams = function () { return self._chatMediaSessionParamsGetter.apply(self, arguments); };

    self.callPrerequisites = ko.pureComputed(function () {
        var clientMedia = self.clientMedia();

        var oppositeUserSession = self.oppositeCallingUserSession();
        if (clientMedia && oppositeUserSession)
        {
            var oppositeCompatibility = GetUserSessionWithCompatibility(oppositeUserSession).co;
            var oppositeLastConnection = oppositeUserSession.User.LastConnection;

            if (oppositeUserSession.IsActive && oppositeCompatibility && oppositeLastConnection)
            {
                let masks = {
                    localMicrophone: 1 << 0,
                    oppositeMicrophone: 1 << 1,
                    audioCompatible: 1 << 2,
                    localCamera: 1 << 0,
                    oppositeCamera: 1 << 1,
                    videoCompatible: 1 << 2,
                    screenAvailable: 1 << 0,
                    localSecuredConnection: 1 << 0,
                    oppositeSecuredConnection: 1 << 1
                };

                var res = {
                    audio: (clientMedia.isMicrophoneAvailable ? 0 : masks.localMicrophone) |
                        (oppositeLastConnection.IsMicrophoneAvailable ? 0 : masks.oppositeMicrophone) |
                        (oppositeCompatibility.isAudioCompatible ? 0 : masks.audioCompatible),
                    video: (clientMedia.isCameraAvailable ? 0 : masks.localCamera) |
                        (oppositeCompatibility.isVideoCompatible ? 0 : masks.videoCompatible),
                    screen: (getScreenOS_Availability(self.userAgent.os.name) && getBrowserScreenAvailability(self.userAgent.browser)) ? 0 : masks.screenAvailable,
                    security: (self.securedConnection ? 0 : masks.localSecuredConnection) |
                        (oppositeLastConnection.SecuredConnection ? 0 : masks.oppositeSecuredConnection),
                    masks: masks
                };

                return res;
            }
        }

        return null;
    }, this);


    self.callOptions = ko.pureComputed(function () {
        var isOthersRinging = false, isOthersCalling = false;

        for (var i = 0; i < self.userSessions().length; i++) {
            var userSession = self.userSessions()[i].us;
            isOthersRinging |= userSession.CallState === chatUtils.ChatUserSessionCallStateEnum.RINGING;
            isOthersCalling |= userSession.CallState === chatUtils.ChatUserSessionCallStateEnum.CALLING;
        }

        var res = {
            canCall: false,
            canPickUp: false,
            canHangUp: false,
            isDialing: false,
            isRinging: false,
            isOthersRinging: isOthersRinging
        };

        var myUserSession = self.myUserSession();
        if (myUserSession)
        {
            res.isRinging = myUserSession.CallState === chatUtils.ChatUserSessionCallStateEnum.RINGING;
            switch (myUserSession.CallState) {
                case chatUtils.ChatUserSessionCallStateEnum.NONE:
                    res.canCall = !isOthersCalling;
                    postData({ callState: chatUtils.ChatUserSessionCallStateEnum.NONE });
                    break;
                case chatUtils.ChatUserSessionCallStateEnum.CALLING:
                    res.canHangUp = myUserSession.CallingConnectionId === $.connection.hub.id;
                    res.isDialing = res.canHangUp && isOthersRinging;
                    postData({ callState: chatUtils.ChatUserSessionCallStateEnum.CALLING });
                    break;
                case chatUtils.ChatUserSessionCallStateEnum.RINGING:
                    res.canPickUp = res.canHangUp = res.isRinging = true;
                    postData({ callState: chatUtils.ChatUserSessionCallStateEnum.RINGING });
                    break;
                default:
                    throw "Unknown call state";
            }
        }

        return res;
    }, this);
    
    function GetUserSessionWithCompatibility(userSession)
    {
        return {
            us: userSession,
            co: userSession.User.LastConnection ? determineClientsWebRTCCompatibility(self.userAgent, userSession.User.LastConnection.UserAgent) : null
        };
    }

    function postData(data) {
        var origin = guestParams && guestParams.freeMode === true ? "*" : window.location !== window.parent.location ? document.referrer : document.location.href;
            
        if (window.parent)
            window.parent.postMessage(data, origin);

        if (window.opener)
            window.opener.postMessage(data, origin);
    }

    self.setChatProfileOperatorsAvailability = function (operatorsAvailability) {
        self.chatProfileOperatorsAvailability.canChatWithOperator(operatorsAvailability.CanReceiveNewGuests);
        self.chatProfileOperatorsAvailability.queuedSessionsCount(operatorsAvailability.QueuedSessionsCount);

        postData({
            canReceiveNewGuests: self.chatProfileOperatorsAvailability.canReceiveNewGuests(),
            chatProfileKey: self.chatProfileKey
        });
    };

    function sendMessageByHub() {
        if (self.messageToSend()) {
            incompleteMessageLength = null;
            chatHub.server.addMessage(self.messageToSend(), sessionId, null).done(function () {
                self.messageToSend("");
            }).fail(function (error) {
                console.error('sendMessage: ' + error);
            });
        }
    }

    function isInGuestChatBotMode() {
        if (self.isGuest && self.chatProfile && self.chatProfile.ChatBot) {
            var session = self.session();
            return !session || session.ChatBotMode;
        }

        return null;
    }

    function setChatBotIncompleteMessageIfNeeded(writingNotErasing) {
        var isInGuestChatBotModeVal = isInGuestChatBotMode();
        if (isInGuestChatBotModeVal === null)
            return;

        if (isInGuestChatBotModeVal) {
            setTimeout(function () {
                var lastMessage = self.messages().length > 0 ? self.messages()[self.messages().length - 1] : null;

                if (lastMessage && lastMessage.UserProviderKey === userProviderKey) {
                    self.addMessage({
                        Text: Localization_Chat_JustWriting + '&hellip;',
                        Type: chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE,
                        IncompleteMessageLength: writingNotErasing ? 1 : 0,
                        UserProviderKey: "",
                        FullName: self.appearance().ChatBot && self.appearance().ChatBot.Name,
                        Date: new Date().getTime(),
                        ChatBot: true,
                        Chosen: false,
                        Options: [],
                        EndChatBotModeOption: false,
                        FromOperator: true
                    });
                }
            }, 500);
        }
        else {
            var lastMessage = self.messages().length > 0 ? self.messages()[self.messages().length - 1] : null;
            if (lastMessage.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
                self.messages.splice(self.messages().length - 1, 1);
            }
        }
    }

    self.setChatBotMode = function (messageId) {
        chatHub.server.setChatBotMode(messageId, sessionId).fail(function (error) {
            console.error('setChatBotMode: ' + error);
        });
    };

    self.sendMessage = function () {
        if (!self.sendingMessage())
        {
            self.sendingMessage(true);
            startSessionIfNeeded().then(sendMessageByHub).then(function () {
                self.sendingMessage(false);
            });
        }
    };

    self.sendOptionsMessage = function (message, messageId) {
        if (!self.sendingMessage()) {
            self.sendingMessage(true);
            chatHub.server.addOptionsMessage(message, messageId, sessionId).done(function () {
                self.sendingMessage(false)
            }).fail(function (error) {
                console.error('sendOptionsMessage: ' + error);
            });
        }
    };

    self.addMessage = function (message) {
        function internalAddMessage() {
            placeMessageAtPosition(message);
            spotifyMessages();

            if (message.UserProviderKey !== userProviderKey) {
                if (message.Type !== chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE)
                    postData({ displayChat: true, sessionId: sessionId });
            } else
                setChatBotIncompleteMessageIfNeeded(true);
        }

        if (isInGuestChatBotMode() === true && self.chatProfile.ChatBot && self.chatProfile.ChatBot.MessageDelay > 0) {
            if (message.FromOperator && message.Type !== chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
                var messageDelayMilSec = self.chatProfile.ChatBot.MessageDelay * 1000;
                var lastMessageSentDiff = new Date() - new Date(self.lastMessageSentDate());
                if (lastMessageSentDiff < messageDelayMilSec) {
                    setTimeout(internalAddMessage, messageDelayMilSec - lastMessageSentDiff);
                    return;
                }
            }
        }

        internalAddMessage();
    };

    function placeMessageAtPosition(message) {
        function delayedRemoveMessageAtPosition(message) {
            setTimeout(function () {
                let messages = self.messages();
                for (let i = messages.length - 1; i >= 0; i--) {
                    let iteratedMessage = messages[i];
                    if (iteratedMessage.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
                        if (iteratedMessage === message) {
                            self.messages.splice(i, 1);
                            break;
                        }
                    }
                    else
                        break;
                }
            }, MAX_EXPECTED_DELAYED_MESSAGE_INTERVAL);
        }

        var justWritingMessage = false;
        if (message.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE && message.Text === null) {
            message.Text = Localization_Chat_JustWriting + '&hellip;';
            justWritingMessage = true;
        }

        for (var i = self.messages().length - 1; i >= 0; i--) {
            var iteratedMessage = self.messages()[i];
            if (iteratedMessage.Type !== chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
                if (message.Type !== chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
                    if (i === length - 1)
                        self.messages.push(message);
                    else
                        self.messages.splice(i + 1, 0, message);

                    return;
                }
            }
            else {
                if (message.UserProviderKey === iteratedMessage.UserProviderKey) {
                    if (message.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
                        if (message.IncompleteMessageLength > 0) {
                            message.writingNotErasing = ko.observable(message.IncompleteMessageLength >= iteratedMessage.IncompleteMessageLength);
                            if (justWritingMessage)
                                delayedRemoveMessageAtPosition(message);
                            self.messages.splice(i, 1, message); //knockout haze chyby pri bindovani, pokud predam undefined jako treti parametr
                        }
                        else
                            self.messages.splice(i, 1);
                        return;
                    }
                    else if (message.Type === chatUtils.ChatMessageTypeEnum.COMMON || message.Type === chatUtils.ChatMessageTypeEnum.ATTACHMENTS ||
                        message.Type === chatUtils.ChatMessageTypeEnum.BOT_COMMON || message.Type === chatUtils.ChatMessageTypeEnum.BOT_OPTIONS) {
                        self.messages.splice(i, 1);
                        continue;
                    }
                }
            }
        }

        if (message.Type === chatUtils.ChatMessageTypeEnum.COMMON_INCOMPLETE) {
            message.writingNotErasing = ko.observable(true);
            if (justWritingMessage)
                delayedRemoveMessageAtPosition(message);
        }

        self.messages.push(message);
    }

    self.terminateSession = function () {
        chatHub.server.terminateSession(sessionId).fail(function (error) {
            console.warn('terminateSession: ' + error);
        });
    };

    self.stopSighting = function () {
        chatHub.server.stopSighting(sessionId).fail(function (error) {
            console.warn('stopSighting: ' + error);
        });
    };

    self.getMyAttitude = function () {
        return self.getAttitudeOfUser(userProviderKey);
    };

    self.getAttitudeOfUser = function (userProviderKey) {
        var expression = self.getExpressionOfUser(userProviderKey);
        return expression ? expression.Attitude : null;
    };

    self.getMyExpression = function () {
        return self.getExpressionOfUser(userProviderKey);
    };

    self.getExpressionOfUser = function (userProviderKey) {
        var session = self.session();
        if (session) {
            let expression = session.Expressions[userProviderKey];
            return expression ? expression : null;
        }

        return null;
    };

    self.isAcceptedByAnybody = ko.pureComputed(function () {
        var session = self.session();
        if (session) {
            for (var prop in session.Expressions) {
                if (session.Expressions.hasOwnProperty(prop)) {
                    if (session.Expressions[prop].Attitude === true)
                        return true;
                }
            }
        }

        return false;
    }, this);

    self.isAcceptedByAnybody.subscribe(function (accepted) {
        postData({ isAcceptedByAnybody: accepted });
    });

    self.acceptSession = function (clearOffer) {
        var session = self.session();
        if (session) {
            chatHub.server.acceptSession(session.SessionId, clearOffer).fail(function (error) {
                console.warn('acceptSession: ' + error);
            });
        }
    };

    var uploader;
    self.initUploader = function (up) {
        up.init();
        up.bind('FilesAdded', function (up, files) {
            self.attachmentsUploadError(null);
            self.attachmentsUpload(files[0]);
            startSessionIfNeeded().then(function () {
                //up.settings.multipart_params = { identifierForAsyncAttachments: guid() }
                up.start();
            });                    
        });

        up.bind('UploadProgress', function (up, file) {
            self.attachmentsUpload(file);
        });

        up.bind('FileUploaded', function (up, file, res) {
            $.ajax({
                url: window.location.protocol + "//" + window.location.host + RQ_URL + "/Chat/GetFileAndSave",
                type: "POST",
                dataType: "json",
                contentType: "application/json; charset=utf-8",
                data: JSON.stringify({ identifierForAsyncAttachments : up.settings.multipart_params.identifierForAsyncAttachments, sessionId: sessionId, userProviderKey: userProviderKey }),
                success: function (g) {
                        
                }
            });
        });

        up.bind('UploadComplete', function (up, files) {
            self.attachmentsUpload(null);
        });

        up.bind('QueueChanged', function (up, files) {
        });

        up.bind('FilesRemoved', function (up, files) {
        });

        up.bind('Error', function (up, err) {
            self.attachmentsUpload(null);
            self.attachmentsUploadError(err);
        });

        uploader = up;

        postData({ uploaderInitialized: true });
    };

    self.stopAttachmentsUpload = function () {
        uploader.stop();
        uploader.splice();
        self.attachmentsUpload(null);
    };

    self.stopConnection = function () {
        if (guestParams) {
            $.connection.hub.stop();
            window.location.reload();
        }
    };

    self.filterHashtags = function (hashtag) {
        if (self.answers()) {
            return $.grep(self.answers(), function (answer) {
                return answer.Shortcut.startsWith(hashtag);
            });
        }

        return [];
    };

    self.getTextOfHashtag = function (hashtag) {
        for (var i = 0; i < self.answers().length; ++i) {
            let answer = self.answers()[i];
            if (answer.Shortcut === hashtag)
                return answer.Text;
        }

        return null;
    };

    function handleWebRTCError(error) {
        console.warn('WebRTC error: ', error);
    }

    function stopTracksOfStream(stream) {
        stream.getTracks().forEach(function (track) {
            track.stop();
        });
    }

    function stopLocalCameraStream() {
        if (self.localCameraStream()) {
            stopTracksOfStream(self.localCameraStream());
            self.localCameraStream(null);
        }

        self.isLocalAudioEnabled(true);
        self.isLocalVideoEnabled(true);
    }

    function restartLocalCameraStream(useCamera) {
        stopLocalCameraStream();
        return navigator.mediaDevices.getUserMedia({
            audio: true,
            video: useCamera
        }).then(function (stream) {
            self.localCameraStream(stream);
        }).catch(handleWebRTCError);
    }

    self.onCall = function (useCamera) {
        if (mediaParams && mediaParams.callMode) {
            restartLocalCameraStream(useCamera).then(function () {
                createPeerConnection();
                let localCameraStream = self.localCameraStream();
                localCameraStream.getTracks().forEach(function (track) {
                    pc.addTrack(track, localCameraStream);
                });

                chatHub.server.callSession(sessionId).fail(function (error) {
                    console.warn('Error callSession: ' + error);
                });
            });
        }
        else {
            window.dispatchEvent(new CustomEvent('ChatOpenMediaWindow', {
                detail: self.chatMediaSessionParams(chatUtils.CallAutomationModeEnum.CALL, useCamera, guestParams)
            }));
        }
    };

    self.onPickUp = function (useCamera) {
        if (mediaParams && mediaParams.pickUpMode) {
            restartLocalCameraStream(useCamera).then(function () {
                createPeerConnection();
                let localCameraStream = self.localCameraStream();
                localCameraStream.getTracks().forEach(function (track) {
                    pc.addTrack(track, localCameraStream);
                });
                handleNegotiation();
            });
        }
        else {
            window.dispatchEvent(new CustomEvent('ChatOpenMediaWindow', {
                detail: self.chatMediaSessionParams(chatUtils.CallAutomationModeEnum.PICK_UP, useCamera, guestParams)
            }));
        }
    };

    function onAnswer(message) {
        if (!self.localCameraStream())
            return;

        pc.setRemoteDescription(new RTCSessionDescription(message));
        pc.createAnswer().then(
             function (sessionDescription) {
                 pc.setLocalDescription(sessionDescription);
                 answerRemoteCall(sessionDescription);
             },
            handleWebRTCError
        );

        function answerRemoteCall(sdpAnswer) {
            chatHub.server.sendCommonMessage(sdpAnswer, sessionId).done(function () {
            }).fail(function (error) {
                console.warn('Error sending answer: ' + error);
            });
        }
    }

    self.leaveSession = function () {
        return new Promise(function (resolve, reject) {
            if (self.isAcceptedByAnybody())
                resolve();

            if (self.isOnServer()) {
                chatHub.server.leaveSession(sessionId).done(function () {
                    resolve(sessionId);
                }).fail(function (error) {
                    console.warn('Error sending leave session: ' + error);
                    reject(error);
                });
            }
            else
                resolve(sessionId);
        });
    };

    self.reenterSession = function () {
        return new Promise(function (resolve, reject) {
            if (self.isOnServer()) {
                chatHub.server.reenterSession(sessionId).done(function () {
                    resolve(sessionId);
                }).fail(function (error) {
                    console.warn('Error sending reenter session: ' + error);
                    reject(error);
                });
            }
            else
                resolve(sessionId);
        });
    };

    self.isLocalAudioEnabled = ko.observable(true);
    self.isLocalVideoEnabled = ko.observable(true);

    self.onToggleAudio = function () {
        let audioTrack = self.localCameraStream().getAudioTracks()[0];
        audioTrack.enabled = !audioTrack.enabled;
        self.isLocalAudioEnabled(audioTrack.enabled);
    };

    self.onToggleVideo = function () {
        let videoTrack = self.localCameraStream().getVideoTracks()[0];
        videoTrack.enabled = !videoTrack.enabled;
        self.isLocalVideoEnabled(videoTrack.enabled);
    };

    self.isScreenSharing = ko.observable(false);

    var screenShareSenders = null;
    self.onToggleScreenShare = function () {
        if (!screenShareSenders) {
            navigator.mediaDevices.getDisplayMedia({
                audio: true,
                video: true
            }).then(function (stream) {
                stream.getTracks().forEach(function (track) {
                    track.onended = function (e) {
                        let track = e.target;
                        let sender = screenShareSenders[track.id];
                        sender.track.onended = null;
                        sender.track.stop();
                        pc.removeTrack(sender);
                        delete screenShareSenders[track.id];
                        if (Object.keys(screenShareSenders).length === 0) {
                            screenShareSenders = null;
                            self.isScreenSharing(false);
                            handleNegotiation();
                        }
                    };

                    screenShareSenders = screenShareSenders || {};
                    screenShareSenders[track.id] = pc.addTrack(track, stream);
                    self.isScreenSharing(true);
                });

                if (!self.callOptions().isDialing)
                    handleNegotiation();
            }).catch(handleWebRTCError);
        }
        else {
            for (let prop in screenShareSenders) {
                let sender = screenShareSenders[prop];
                sender.track.onended = null;
                sender.track.stop();
                pc.removeTrack(sender);
            }
            
            screenShareSenders = null;
            self.isScreenSharing(false);
            handleNegotiation();
        }
    };

    self.onHangUp = function () {
        disposeRemoteCall();
        stopLocalCameraStream();
        hangUpAndFinishRemoteCall(sessionId);

        function hangUpAndFinishRemoteCall(sessionId) {
            chatHub.server.hangUpAndFinishRemoteCall(sessionId).fail(function (error) {
                console.warn('Error sending hangUp: ' + error);
            });
        }
    };

    function handleNegotiation() {
        pc.createOffer({
            offerToReceiveAudio: 1,
            offerToReceiveVideo: 1,
            voiceActivityDetection: false
        }).then(
            function (sessionDescription) {
                //sessionDescription.sdp = preferOpus(sessionDescription.sdp);
                pc.setLocalDescription(sessionDescription);

                if (self.callOptions().isRinging)
                    pickUpAndOfferRemoteCall(sessionDescription);
                else if (self.callOptions().canHangUp)
                    sendCommonMessage(sessionDescription);
            },
            handleWebRTCError
        );

        function pickUpAndOfferRemoteCall(sdpOffer) {
            chatHub.server.pickUpAndOfferRemoteCall(sdpOffer, sessionId).done(function () {
            }).fail(function (error) {
                console.warn('Error sending offer: ' + error);
            });
        }
    }

    function createPeerConnection() {
        try {
            var conf = {
                'iceServers': [
                  {
                      'urls': 'stun:stun.l.google.com:19302'
                  },
                  {
                      'urls': "turn:turn.anyfirewall.com:443?transport=tcp",
                      'credential': 'webrtc',
                      'username': 'webrtc'
                  },
                  {
                      'urls': 'turn:numb.viagenie.ca',
                      'credential': 'requestor',
                      'username': 'lukas.vyslouzil@requestor.com'
                  }
                ]
            };

            pc = new RTCPeerConnection(conf);
            pc.onicecandidate = handleIceCandidate;
            pc.ontrack = handleRemoteStreamAdded;
            pc.onsignalingstatechange = handleSignalingStateChange;
        } catch (e) {
            handleWebRTCError(e.message);
            return;
        }

        function handleSignalingStateChange() {
            removeInactiveRemoteStreams();
        }

        function removeInactiveRemoteStreams() {
            let streamObservables = [self.remoteCameraStream, self.remoteScreenStream];
            streamObservables.forEach(function (streamObservable) {
                let stream = streamObservable();
                if (stream && !stream.active) {
                    stopTracksOfStream(stream);
                    streamObservable(null);
                }
            });
        }

        function handleRemoteStreamAdded(event) {
            event.streams.forEach(function (stream) {
                if (!self.remoteCameraStream())
                    self.remoteCameraStream(stream);
                else if (!self.remoteScreenStream() && self.remoteCameraStream().id !== stream.id)
                    self.remoteScreenStream(stream);
            });
        }

        function handleIceCandidate(event) {
            if (event.candidate) {
                sendCommonMessage({
                    type: 'candidate',
                    label: event.candidate.sdpMLineIndex,
                    id: event.candidate.sdpMid,
                    candidate: event.candidate.candidate
                });
            }
        }
    }

    function sendCommonMessage(message) {
        chatHub.server.sendCommonMessage(message, sessionId).done(function () {
        }).fail(function (error) {
            console.warn('Error sending common message: ' + error);
        });
    }

    self.receiveCommonMessage = function (msgEnvelope) {        
        if (msgEnvelope.sessionId !== sessionId)
            return;

        var message = msgEnvelope.msg;
        if (message.type === 'offer') {
            onAnswer(message);
        }
        else if (message.type === 'answer') {
            if (pc !== null)
                pc.setRemoteDescription(new RTCSessionDescription(message));
        }
        else if (message.type === 'candidate') {
            if (pc !== null) {
                var candidate = new RTCIceCandidate({
                    sdpMLineIndex: message.label,
                    candidate: message.candidate
                });
                pc.addIceCandidate(candidate);
            }
        } else if (message === 'finish') {
            disposeRemoteCall();
            stopLocalCameraStream();
        }
    };

    function disposeRemoteCall() {
        if (self.remoteCameraStream())
        {
            stopTracksOfStream(self.remoteCameraStream());
            self.remoteCameraStream(null);
        }
        disposePeerConnection();
    }

    function disposePeerConnection() {
        if (pc !== null) {
            pc.close();
            pc = null;
        }
    }

    function initConnectionHub() {
        $.connection.hub.qs = {
            'userProviderKey': userProviderKey,
            'sessionId': sessionId,
            'userAgent': JSON.stringify(self.userAgent),
            'clientMedia': self.clientMedia(),
            'hostUrl': guestParams && guestParams.hostUrl,
            'hostDocumentTitle': guestParams && guestParams.hostDocumentTitle,
            'guestChatProfileKey': guestParams && guestParams.chatProfileKey,
            'preferContextUser': guestParams && guestParams.preferContextUser,
            'language': navigator.language || navigator.userLanguage,
            'geo': geoStr,
            'securedConnection': self.securedConnection,
            'clientMessagesScaffolds': clientMessagesScaffolds ? JSON.stringify(clientMessagesScaffolds) : null,
            'clientEpochTime': new Date().getTime()
        };

        $.connection.hub.reconnecting(function () {
            self.isConnected(false);
        });

        $.connection.hub.reconnected(function () {
            self.isConnected(true);
        });

        $.connection.hub.disconnected(function () {
            self.isConnected(false);
            connectAfterTimeout();
        });

        chatHub = $.connection.chatHub;
        chatHub.client.setUserSessions = self.setUserSessions;
        chatHub.client.addMessage = self.addMessage;
        chatHub.client.setChatProfileOperatorsAvailability = self.setChatProfileOperatorsAvailability;
        chatHub.client.stopConnection = self.stopConnection;
        chatHub.client.receiveCommonMessage = self.receiveCommonMessage;
    }

    self.visibleHandler = function () {
        if (!chatHub)
            chatSessionStateCheck();
    };
    self.notVisibleHandler = function () {
        if (!chatHub)
            clearTimeout(chatSessionStateTimeout);
    };

    var geoStr, texts;
    self.init = function (userSessions, passedChatHub) {
        self.formEditMode(guestParams ? true : false);

        if (!passedChatHub) {
            window.addEventListener("beforeunload", function (event) {
                $.connection.hub.stop();
            });
        }
        else
            chatHub = passedChatHub;

        geoStr = guestParams ? chatUtils.loadGeoFromStorage() : null;
        return new Promise(function (resolve, reject) {
            chatUtils.detectClientMedia().then(function (clientMedia) {
                self.clientMedia(clientMedia);
                if (userSessions)
                    self.setUserSessions(userSessions);

                loadChatProfileData().then(function (e) {
                    if (self.guestUserSession())
                        fillFormValues(self.guestUserSession().Session.Form);

                    if (guestParams)
                    {
                        texts = e.Texts;
                        sessionStorage = chatUtils.getSessionStorage(self.chatProfileKey);
                        postData({ sessionStorage: sessionStorage });
                        if (Array.isArray(e.Automations) && e.Automations.length > 0)
                            startAutomationTimer(e.Automations);

                        if (!chatHub) {
                            chatSessionStateCheck();
                        }
                        else {
                            chatHub.server.getChatProfileOperatorsAvailability(sessionId, self.chatProfileKey).done(function (operatorsAvailability) {
                                self.setChatProfileOperatorsAvailability(operatorsAvailability);
                            }).fail(function (error) {
                                console.error('getChatProfileOperatorsAvailability: ' + error);
                            });
                        }
                    }
                    else
                    {
                        if (mediaParams && !chatHub)
                        {
                            initConnectionHub();
                            connect();
                        }
                    }

                }).catch(function (error) {

                });
                resolve();
            }).catch(function (error) {
                reject(error);
            });
        });
    };

    var chatSessionStateTimeout;
    function chatSessionStateCheck() {
        clearTimeout(chatSessionStateTimeout);
        
        if (chatHub/*$.connection.hub.state !== $.signalR.connectionState.disconnected*/)
            return;

        $.getJSON((window.location.protocol + '//' + window.location.host + RQ_URL + "/Chat/GetChatSessionState/" + sessionId + "?chatProfileKey=" + self.chatProfileKey)).done(function (res) {
            if (res.Started) {
                postData({ displayChat: true, sessionId: sessionId });
                if (!chatHub)
                    initConnectionHub();

                connect();
            }
            else {
                self.setChatProfileOperatorsAvailability(res.OperatorsAvailability);
                chatSessionStateTimeout = setTimeout(chatSessionStateCheck, REPEAT_SESSION_CHECK);
            }
        }).fail(function () {
            chatSessionStateTimeout = setTimeout(chatSessionStateCheck, REPEAT_SESSION_CHECK);
        });
    }

    function connect() {
        if ($.connection.hub.state === $.signalR.connectionState.disconnected) {
            $.connection.hub.start().done(connectionDone);
        }
    }

    function connectAfterTimeout() {
        setTimeout(function () {
            connect();
        }, REPEAT_TIMEOUT);
    }

    function connectionDoneAfterTimeout() {
        setTimeout(function () {
            if ($.connection.hub.state !== $.signalR.connectionState.disconnected) {
                connectionDone();
            }
        }, REPEAT_TIMEOUT);
    }

    function processAutomationModeIfNeeded() {
        if (mediaParams)
        {
            if (mediaParams.pickUpMode) {
                if (mediaParams.pickUpMode === "mic")
                    self.onPickUp(false);
                else if (mediaParams.pickUpMode === "cam")
                    self.onPickUp(true);
            }
            else if (mediaParams.callMode) {
                if (mediaParams.callMode === "mic")
                    self.onCall(false);
                else if (mediaParams.callMode === "cam")
                    self.onCall(true);
            }
        }
    }

    function connectionDone() {
        return new Promise(function (resolve, reject) {
            self.isConnected(true);
            if (chatProfileDataProcessed) {
                initUserSessions().then(function (e) {
                    processAutomationModeIfNeeded();
                    resolve(e);
                }).catch(function (err) {
                    connectionDoneAfterTimeout();
                    reject(err);
                });
            }
            else {
                initChatProfileDataAndUserSessions().then(function (e) {
                    processAutomationModeIfNeeded();
                    resolve(e);
                }).catch(function (err) {
                    connectionDoneAfterTimeout();
                    reject(err);
                });
            }
        });
    }

    function initUserSessions() {
        self.isLoading(true);
        return new Promise(function (resolve, reject) {
            chatHub.server.getUserSessionsOfSession(sessionId).done(function (userSessions) {
                self.setUserSessions(userSessions);
                if (self.myUserSession().Session.GuestProviderKey === self.myUserSession().User.UserProviderKey) {
                    chatHub.server.getChatProfileOperatorsAvailability(sessionId, self.chatProfileKey).done(function (operatorsAvailability) {
                        self.setChatProfileOperatorsAvailability(operatorsAvailability);
                    }).fail(function (error) {
                        console.warn('getChatProfileOperatorsAvailability: ' + error);
                    });
                }
                else {
                    self.setChatProfileOperatorsAvailability({
                        CanReceiveNewGuests: true,
                        QueuedSessionsCount: null
                    });
                }

                self.isLoading(false);
                resolve();
            }).fail(function (error) {
                self.isLoading(false);
                initUserSessionsAfterTimeout();
                reject(error);
            });
        });
    }

    function initChatProfileDataAndUserSessions() {
        self.isLoading(true);
        return new Promise(function (resolve, reject) {
            loadChatProfileData().then(function (e) {
                initUserSessions().then(function (e) {
                    resolve(e);
                }).catch(function (err) {
                    reject(err);
                });
            }).catch(function (error) {
                self.isLoading(false);
                setTimeout(function () {
                    initChatProfileDataAndUserSessions().then(function (e) {
                        resolve(e);
                    }).catch(function (err) {
                        reject(err);
                    });
                }, REPEAT_TIMEOUT);
            });
        });
    }

    function initUserSessionsAfterTimeout() {
        setTimeout(function () {
            initUserSessions();
        }, REPEAT_TIMEOUT);
    }
}

ChatSessionViewModel.prototype._oppositeCallingUserSessionGetter = function () {
    var self = this;

    var oppositeUserSession;
    for (var i = 0; i < self.userSessions().length; i++) {
        if (self.userSessions()[i].us.CallState === chatUtils.ChatUserSessionCallStateEnum.CALLING) {
            oppositeUserSession = self.userSessions()[i].us;
            break;
        }
    }

    if (!oppositeUserSession)
        oppositeUserSession = self.guestUserSession();

    return oppositeUserSession;
};

ChatSessionViewModel.prototype._chatMediaSessionParamsGetter = function (automationMode, useCamera, guestParams) {
    var self = this;

    return {
        chatProfileKey: self.chatProfileKey,
        pickUpMode: automationMode === chatUtils.CallAutomationModeEnum.PICK_UP ? useCamera ? "cam" : "mic" : null,
        callMode: automationMode === chatUtils.CallAutomationModeEnum.CALL ? useCamera ? "cam" : "mic" : null,
        sessionId: self.sessionId,
        guestMode: self.isGuest,
        preferContextUser: guestParams && guestParams.preferContextUser ? true : false
    };
};


